import anyio
import hashlib
import hmac
import secrets
import typing as t
from starlette.requests import HTTPConnection

from .exceptions import SessionReusageError
from .protocols import LoginGuard, PasswordVerifier, UserLike
from .user_providers import UserProvider
from .user_token import AnonymousUser, LoginState, UserToken

SESSION_KEY = '_auth_user_id'
SESSION_HASH = '_auth_user_hash'
HASHING_ALGORITHM = 'sha1'


def get_session_auth_id(connection: HTTPConnection) -> t.Optional[str]:
    return connection.session.get(SESSION_KEY)


def get_session_auth_hash(connection: HTTPConnection) -> t.Optional[str]:
    return connection.session.get(SESSION_HASH)


def _get_password_hmac_from_user(user: UserLike, secret: str) -> str:
    """Generate HMAC value for user's password."""
    key = hashlib.sha256(("imia.session.hash" + secret).encode()).digest()
    return hmac.new(key, msg=user.get_hashed_password().encode(), digestmod=hashlib.sha1).hexdigest()


def _check_for_other_user_session(connection: HTTPConnection, user: UserLike, user_password_hmac: str) -> None:
    """
    There is a chance that session may already contain data of another user.
    This may happen if you don't clear session property on logout, or
    SESSION_KEY is set from the outside. In this case we need to run several
    security checks to ensure that SESSION_KEY is valid.

    Our plan:
        * if SESSION_KEY and ID of current user is not the same -> risk of session re-usage
        * if session hash is not the same as hash for user's password then we clearly reusing other's session.
    """
    session_auth_hmac = get_session_auth_hash(connection)
    if SESSION_KEY in connection.session and any(
        [
            # if we have other user id in the session
            connection.session[SESSION_KEY] != str(user.get_id()),
            # and session has previously set hash, and hashes are not equal
            session_auth_hmac and not secrets.compare_digest(session_auth_hmac, user_password_hmac),
        ]
    ):
        # probably this is the session of another user -> clear
        raise SessionReusageError()


async def login_user(request: HTTPConnection, user: UserLike, secret_key: str) -> UserToken:
    """Login a user w/o password check."""
    user_password_hmac = _get_password_hmac_from_user(user, secret_key)
    try:
        _check_for_other_user_session(request, user, user_password_hmac)
    except SessionReusageError:
        request.session.clear()
    else:
        if hasattr(request.session, 'regenerate_id'):
            # if session implements `def regenerate_id(self) -> str` then call it
            await request.session.regenerate_id()  # type: ignore

    user_token = UserToken(user=user, state=LoginState.FRESH)
    request.session[SESSION_KEY] = user.get_id()
    request.session[SESSION_HASH] = user_password_hmac
    request.scope['auth'] = user_token
    return user_token


class LoginManager:
    """Use this class to handle login and logout forms."""

    def __init__(self, user_provider: UserProvider, password_verifier: PasswordVerifier, secret_key: str = '') -> None:
        self._user_provider = user_provider
        self._password_verifier = password_verifier
        self._secret_key = secret_key

    async def login(
        self, request: HTTPConnection, username: str, password: str, guards: t.Optional[t.Iterable[LoginGuard]] = None
    ) -> UserToken:
        user = await self._user_provider.find_by_username(username)
        guards = guards or []
        async with anyio.create_task_group() as tg:
            for guard in guards:
                tg.start_soon(guard, request, user)

        if user is not None and self._password_verifier.verify(password, user.get_hashed_password()):
            return await login_user(request, user, self._secret_key)
        return UserToken(user=AnonymousUser(), state=LoginState.ANONYMOUS)

    async def logout(self, request: HTTPConnection) -> None:
        request.session.clear()
        if hasattr(request.session, 'regenerate_id'):
            await request.session.regenerate_id()  # type: ignore

        request.scope['auth'] = UserToken(user=AnonymousUser(), state=LoginState.ANONYMOUS)



import typing as t
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Receive, Scope, Send

from .exceptions import AuthenticationError
from .protocols import UserLike
from .user_providers import UserProvider
from .user_token import LoginState, UserToken

IMPERSONATION_SESSION_KEY = '_impersonated_user_id'


class ImpersonationNotAllowedError(AuthenticationError):
    """Raised when the user is not allowed to perform impersonation."""


class ImpersonationNotActiveError(AuthenticationError):
    """Raised when you try to access impersonation related data but the
    impersonation is not active."""


def impersonate(request: HTTPConnection, user: UserLike) -> None:
    """Activate impersonation."""
    request.scope['auth'] = UserToken(user, state=LoginState.IMPERSONATOR, original_user_token=request.scope['auth'])
    if 'session' in request.scope:
        request.scope['session'][IMPERSONATION_SESSION_KEY] = user.get_id()


def exit_impersonation(request: HTTPConnection) -> None:
    """Exit the impersonation session (restores to an original user)."""
    if 'session' in request.scope:
        request.scope['session'].pop(IMPERSONATION_SESSION_KEY, None)


def impersonation_is_active(request: HTTPConnection) -> bool:
    return request.scope['auth'].original_user_id is not None


def get_original_user(request: HTTPConnection) -> UserLike:
    """Get the original user when the impersonation is active."""
    return (
        request.scope['auth'].original_user_token.user
        if request.scope['auth'].original_user_token
        else request.scope['auth'].user
    )


class ImpersonationMiddleware:
    """A middleware used to temporary impersonate another user."""

    def __init__(
        self,
        app: ASGIApp,
        user_provider: UserProvider,
        guard_fn: t.Callable[[UserToken, HTTPConnection], bool] = None,
        enter_query_param: str = "_impersonate",
        exit_user_name: str = "__exit__",
        scope: str = 'auth:impersonate_others',
    ) -> None:
        self._app = app
        self._user_provider = user_provider
        self._guard_fn = guard_fn
        self._enter_query_param = enter_query_param
        self._exit_user_name = exit_user_name
        self._scope = scope

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ["http", "websocket"]:  # pragma: no cover
            await self._app(scope, receive, send)
            return

        request = HTTPConnection(scope)
        action, impersonation_target = self._detect_action(request)
        if action == 'ignore':
            # user haven't asked anything we can offer
            return await self._app(scope, receive, send)

        if 'auth' not in request.scope:
            raise ValueError('ImpersonationMiddleware needs AuthenticationMiddleware to be installed.')

        if not self._can_enter_impersonation(request):
            return await self._app(scope, receive, send)

        if action == 'enter':
            user_id = request.query_params[self._enter_query_param]
            await self._enter_impersonation(request, user_id)

        if action == 'exit':
            await self._exit_impersonation(request)

        if action == 'activate':
            await self._enter_impersonation(request, impersonation_target)

        await self._app(scope, receive, send)

    async def _enter_impersonation(self, request: HTTPConnection, user_id: str) -> None:
        user = await self._user_provider.find_by_id(user_id)
        if user:
            impersonate(request, user)

    async def _exit_impersonation(self, request: HTTPConnection) -> None:
        exit_impersonation(request)

    def _can_enter_impersonation(self, request: HTTPConnection) -> bool:
        """
        Test if current user can impersonate other.

        Here are two checks. The first one to lookup a presence of self._scope
        in token scopes. The other one is to provide guard functions that must
        return boolean value. The guard function take the precedence when
        available.
        """
        if self._guard_fn:
            # forbid impersonation if guard function returns False
            return self._guard_fn(request.auth, request)

        # user must have "can_impersonate" scope
        return self._scope in request.auth

    def _detect_action(self, request: HTTPConnection) -> t.Tuple[str, str]:
        username = request.query_params.get(self._enter_query_param)
        if username is None:
            impersonation_target = request.scope.get('session', {}).get(IMPERSONATION_SESSION_KEY)
            if impersonation_target is not None:
                return 'activate', impersonation_target
            return 'ignore', ''

        if username == self._exit_user_name:
            return 'exit', ''

        return 'enter', username



class AuthenticationError(Exception):
    """Base class for all authentication related errors."""


class SessionReusageError(AuthenticationError):
    """Raise when another user tries to reuse other user session."""



from __future__ import annotations

import abc
import base64
import re
import typing as t
from starlette.requests import HTTPConnection
from starlette.responses import RedirectResponse, Response
from starlette.types import ASGIApp, Receive, Scope, Send

from .exceptions import AuthenticationError
from .login import get_session_auth_id
from .protocols import PasswordVerifier, UserLike
from .user_providers import UserProvider
from .user_token import AnonymousUser, LoginState, UserToken


class WWWAuthenticationRequiredError(Exception):
    pass


class BaseAuthenticator(abc.ABC):  # pragma: no cover
    """
    Authenticators load user using request data. For example, an authenticate
    may use session to get user's ID and load user instance from a user
    provider. Another example can be when you load a user by API token read from
    headers.

    Authenticators are part of AuthenticationMiddleware.
    """

    async def authenticate(self, connection: HTTPConnection) -> t.Optional[UserLike]:
        raise NotImplementedError()

    def get_auth_header(self) -> t.Optional[str]:
        return None


class HTTPBasicAuthenticator(BaseAuthenticator):
    """Basic authenticator supports WWW-Basic authentication type."""

    def __init__(
        self,
        user_provider: UserProvider,
        password_verifier: PasswordVerifier,
        realm: str = 'Protected access',
    ) -> None:
        self.user_provider = user_provider
        self.password_verifier = password_verifier
        self.realm = realm

    async def authenticate(self, connection: HTTPConnection) -> t.Optional[UserLike]:
        header = connection.headers.get('authorization')
        if not header or not header.lower().startswith('basic'):
            raise WWWAuthenticationRequiredError()

        try:
            username, password = base64.b64decode(header[6:]).decode().split(':')
            if password == '':
                raise ValueError('Empty password.')
        except ValueError:
            return None

        user = await self.user_provider.find_by_username(username)
        if not user:
            return None

        if self.password_verifier.verify(password, user.get_hashed_password()):
            return user

        return None

    def get_auth_header(self) -> t.Optional[str]:
        return 'Basic realm="%s"' % self.realm


class SessionAuthenticator(BaseAuthenticator):
    """Session authenticator will use session to get an ID of a user."""

    def __init__(self, user_provider: UserProvider) -> None:
        self.user_provider = user_provider

    async def authenticate(self, connection: HTTPConnection) -> t.Optional[UserLike]:
        user_id = get_session_auth_id(connection)
        if user_id is None:
            return None

        return await self.user_provider.find_by_id(user_id)


class TokenAuthenticator(BaseAuthenticator):
    """Token authenticator reads Authorization header to obtain an API token and
    load a user using it."""

    def __init__(self, user_provider: UserProvider, token_name: str) -> None:
        self.user_provider = user_provider
        self.token_name = token_name

    async def authenticate(self, connection: HTTPConnection) -> t.Optional[UserLike]:
        header = connection.headers.get('authorization')
        if not header:
            return None
        try:
            token_name, token_value = header.split(' ')
        except ValueError:
            return None
        else:
            if token_name.lower() != self.token_name.lower():
                return None
            return await self.user_provider.find_by_token(token_value)


class BearerAuthenticator(TokenAuthenticator):
    """Bearer authenticator is a subtype of TokenAuthenticator designed for
    Bearer token types."""

    def __init__(self, user_provider: UserProvider) -> None:
        super().__init__(user_provider, 'Bearer')


class APIKeyAuthenticator(BaseAuthenticator):
    """
    API key is a simple way to use token authentication.

    The basic principle is to read token from query params, and fallback to
    headers if none found.
    """

    def __init__(self, user_provider: UserProvider, query_param: str = 'apikey', header_name: str = 'X-Api-Key'):
        self.user_provider = user_provider
        self.query_param = query_param
        self.header_name = header_name

    async def authenticate(self, connection: HTTPConnection) -> t.Optional[UserLike]:
        token = self._get_token_from_query_params(connection)
        token = token or self._get_token_from_header(connection)
        if token:
            return await self.user_provider.find_by_token(token)
        return None

    def _get_token_from_query_params(self, connection: HTTPConnection) -> t.Optional[str]:
        return connection.query_params.get(self.query_param)

    def _get_token_from_header(self, connection: HTTPConnection) -> t.Optional[str]:
        return connection.headers.get(self.header_name)


class AuthenticationMiddleware:
    """
    Authenticator middleware will load a user from the request using
    authenticators.

    If the user can be found Request.auth property will become available.
    """

    def __init__(
        self,
        app: ASGIApp,
        authenticators: t.List[BaseAuthenticator],
        on_failure: str = "do_nothing",  # one of: raise, redirect, do_nothing
        redirect_to: str = "/",
        exclude_patterns: t.List[t.Union[str, t.Pattern]] = None,
        include_patterns: t.List[t.Union[str, t.Pattern]] = None,
    ) -> None:
        if on_failure == 'redirect' and redirect_to is None:
            raise ValueError(
                'redirect_to attribute of AuthenticationMiddleware cannot be None '
                'if on_failure is set to "redirect".'
            )

        if exclude_patterns is not None and include_patterns is not None:
            raise ValueError('"exclude_patterns" and "include_patterns" are mutially exclusive.')

        self._app = app
        self._authenticators = authenticators
        self._on_failure = on_failure
        self._redirect_to = redirect_to
        self._exclude_patterns = exclude_patterns or []
        self._include_patterns = include_patterns

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ["http", "websocket"]:  # pragma: no cover
            await self._app(scope, receive, send)
            return

        # always populate scope['auth']
        scope['auth'] = UserToken(AnonymousUser(), LoginState.ANONYMOUS)

        request = HTTPConnection(scope)
        if self._should_interrupt(request):
            return await self._app(scope, receive, send)

        user: t.Optional[UserLike] = None
        for authenticator in self._authenticators:
            try:
                user = await authenticator.authenticate(request)
                if user:
                    break
            except WWWAuthenticationRequiredError:
                auth_header = authenticator.get_auth_header()
                if auth_header:
                    response = Response(None, headers={'WWW-Authenticate': auth_header}, status_code=401)
                    return await response(scope, receive, send)

        if user:
            scope['auth'] = UserToken(user=user, state=LoginState.FRESH)
        elif self._on_failure == 'raise':
            raise AuthenticationError('Could not authenticate request.')
        elif self._on_failure == 'redirect':
            response = RedirectResponse(self._redirect_to)
            return await response(scope, receive, send)
        elif self._on_failure != 'do_nothing':
            raise ValueError(
                'Unsupported action passed to AuthenticationMiddleware via on_failure argument: '
                '%s.' % self._on_failure
            )
        await self._app(scope, receive, send)

    def _should_interrupt(self, request: HTTPConnection) -> bool:
        for pattern in self._exclude_patterns:
            if re.search(pattern, str(request.url)):
                return True

        return bool(
            self._include_patterns
            and not any(re.search(pattern, str(request.url)) for pattern in self._include_patterns)
        )



import sqlalchemy as sa
import typing as t
from sqlalchemy import select
from sqlalchemy.engine import Result
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncSession
from sqlalchemy.orm import DeclarativeMeta, sessionmaker

from imia import UserLike, UserProvider


class SQLAlchemyCoreUserProvider(UserProvider):
    def __init__(
        self,
        engine: AsyncEngine,
        user_model: type,
        identity_table: sa.Table,
        identity_column: str = 'id',
        username_column: str = 'email',
        api_token_column: str = 'api_token',
    ) -> None:
        self._engine = engine
        self._user_model = user_model
        self._identity_table = identity_table
        self._identity_column = identity_column
        self._username_column = username_column
        self._api_token_column = api_token_column

    async def find_by_id(self, identifier: t.Any) -> t.Optional[UserLike]:
        return await self._fetch_user_by_column(self._identity_column, identifier)

    async def find_by_username(self, username_or_email: str) -> t.Optional[UserLike]:
        return await self._fetch_user_by_column(self._username_column, username_or_email)

    async def find_by_token(self, token: str) -> t.Optional[UserLike]:
        return await self._fetch_user_by_column(self._api_token_column, token)

    async def _fetch_user_by_column(self, column: str, value: t.Any) -> t.Optional[UserLike]:
        column = getattr(self._identity_table.c, column)
        stmt = select(self._identity_table).where(column == value)
        async with self._engine.begin() as connection:  # type: AsyncConnection
            result: Result = await connection.execute(stmt)
        row = result.one_or_none()
        if row:
            return self._user_model(**dict(row))
        return None


class SQLAlchemyORMUserProvider(UserProvider):
    def __init__(
        self,
        session_maker: sessionmaker,
        user_model: DeclarativeMeta,
        identity_column: str = 'id',
        username_column: str = 'email',
        api_token_column: str = 'api_token',
    ) -> None:
        self._session_maker = session_maker
        self._user_model = user_model
        self._identity_column = identity_column
        self._username_column = username_column
        self._api_token_column = api_token_column

    async def find_by_id(self, identifier: t.Any) -> t.Optional[UserLike]:
        return await self._fetch_user_by_column(self._identity_column, identifier)

    async def find_by_username(self, username_or_email: str) -> t.Optional[UserLike]:
        return await self._fetch_user_by_column(self._username_column, username_or_email)

    async def find_by_token(self, token: str) -> t.Optional[UserLike]:
        return await self._fetch_user_by_column(self._api_token_column, token)

    async def _fetch_user_by_column(self, column: str, value: t.Any) -> t.Optional[UserLike]:
        column = getattr(self._user_model, column)
        stmt = select(self._user_model).where(column == value)
        async with self._session_maker() as session:  # type: AsyncSession
            result: Result = await session.execute(stmt)
        return result.scalar_one_or_none()



import typing as t
from starlette.requests import HTTPConnection, Request


class PasswordVerifier(t.Protocol):  # pragma: no cover
    def verify(self, plain: str, hashed: str) -> bool:
        ...


class HasDisplayName(t.Protocol):  # pragma: no cover
    def get_display_name(self) -> str:
        ...


class HasId(t.Protocol):  # pragma: no cover
    def get_id(self) -> t.Any:
        ...


class HasHashedPassoword(t.Protocol):  # pragma: no cover
    def get_hashed_password(self) -> str:
        ...


class HasScopes(t.Protocol):  # pragma: no cover
    def get_scopes(self) -> t.List[str]:
        ...


class UserLike(HasId, HasScopes, HasHashedPassoword, HasDisplayName, t.Protocol):  # pragma: no cover
    ...


class LoginGuard(t.Protocol):
    async def __call__(self, request: Request, user: UserLike) -> None:
        ...


class Authenticator(t.Protocol):  # pragma: no cover_
    async def authenticate(self, connection: HTTPConnection) -> t.Optional[UserLike]:
        ...

    def get_auth_header(self) -> t.Optional[str]:
        ...



from __future__ import annotations

import enum
import typing as t

from .protocols import UserLike


class LoginState(enum.Enum):
    ANONYMOUS = "ANONYMOUS"
    IMPERSONATOR = "IMPERSONATOR"
    REMEMBERED = "REMEMBERED"
    FRESH = "FRESH"


class UserToken:
    __slots__ = ["_user", "_state", 'original_user_token']

    def __init__(
        self,
        user: UserLike,
        state: LoginState,
        original_user_token: UserToken = None,
    ) -> None:
        self._user = user
        self._state = state
        self.original_user_token = original_user_token

    @property
    def is_authenticated(self) -> bool:
        """
        Get authentication state.

        Returns True is user is authenticated.
        """
        return self.state != LoginState.ANONYMOUS

    @property
    def is_anonymous(self) -> bool:
        """Test if current user is not authenticated (anonymous)."""
        return not self.is_authenticated

    @property
    def original_user_id(self) -> t.Optional[t.Any]:
        """Get ID of user being impersonated."""
        return self.original_user_token.user.get_id() if self.original_user_token else None

    @property
    def scopes(self) -> t.List[str]:
        """
        Return permission scopes of current user.

        Returns an empty list for unauthenticated user.
        """
        return self.user.get_scopes()

    @property
    def user_id(self) -> t.Any:
        """Get ID of current user."""
        return self.user.get_id()

    @property
    def user(self) -> UserLike:
        """Get an user model associated."""
        return self._user

    @property
    def display_name(self) -> str:
        """Get a display name of current user."""
        return self.user.get_display_name()

    @property
    def state(self) -> LoginState:
        """Get a login state."""
        return self._state

    def __bool__(self) -> bool:
        """Support if-like usage:
        if user_token:
            ...
        """
        return self.is_authenticated

    def __str__(self) -> str:
        return self.display_name

    def __contains__(self, permission: str) -> bool:
        """Check if permission is in scopes."""
        return permission in self.scopes


class AnonymousUser:
    def get_display_name(self) -> str:
        return 'Anonymous'

    def get_id(self) -> t.Any:
        return None

    def get_hashed_password(self) -> str:
        return ''

    def get_scopes(self) -> t.List[str]:
        return []



import abc
import typing as t

from .protocols import UserLike


class UserProvider(abc.ABC):  # pragma: no cover
    """
    User provides perform user look ups over data storages.

    These classes are consumed by Authenticator instances and are not designed
    to be a part of login or logout process.
    """

    async def find_by_id(self, identifier: t.Any) -> t.Optional[UserLike]:
        """Look up a user by ID."""
        raise NotImplementedError()

    async def find_by_username(self, username_or_email: str) -> t.Optional[UserLike]:
        """
        Look up a user by it's identity.

        Where identity may be an email address, or username.
        """
        raise NotImplementedError()

    async def find_by_token(self, token: str) -> t.Optional[UserLike]:
        """Look up a user using API token."""
        raise NotImplementedError()


class InMemoryProvider(UserProvider):
    """A user provides that uses a predefined map of users."""

    def __init__(self, user_map: t.Mapping[t.Any, UserLike]) -> None:
        self.user_map = user_map

    async def find_by_id(self, identifier: str) -> t.Optional[UserLike]:
        return self.user_map.get(identifier)

    async def find_by_username(self, username_or_email: str) -> t.Optional[UserLike]:
        return self.user_map.get(username_or_email)

    async def find_by_token(self, token: str) -> t.Optional[UserLike]:
        return self.user_map.get(token)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imia', 'imia.ext']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'imia',
    'version': '0.4.0',
    'description': 'Full stack authentication library for ASGI.',
    'long_description': '# Imia\n\nImia (belarussian for "a name") is an authentication library for Starlette and FastAPI (python 3.8+).\n\n![PyPI](https://img.shields.io/pypi/v/imia)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/alex-oleshkevich/imia/Lint)\n![GitHub](https://img.shields.io/github/license/alex-oleshkevich/imia)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/imia)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/imia)\n![GitHub Release Date](https://img.shields.io/github/release-date/alex-oleshkevich/imia)\n![Lines of code](https://img.shields.io/tokei/lines/github/alex-oleshkevich/imia)\n\n## Production status\n\nThe library is considered in "beta" state thus may contain bugs or security issues, but I actively use it in production.\n\n## Installation\n\nInstall `imia` using PIP or poetry:\n\n```bash\npip install imia\n# or\npoetry add imia\n```\n\n## Features\n\n- Login/logout flows\n- Pluggable authenticators:\n    - WWW-Basic\n    - session\n    - token\n    - bearer token\n    - any token (customizable)\n    - API key\n- Database agnostic user storage\n- Authentication middleware\n    - with fallback strategies:\n        - redirect to an URL\n        - raise an exception\n        - do nothing\n    - with optional URL protection\n    - with option URL exclusion from protection\n- User Impersonation (stateless and stateful)\n- SQLAlchemy 1.4 (async mode) integration\n\n## TODO\n\n* remember me\n\n## A very quick start\n\nIf you are too lazy to read this doc, take a look into `examples/` directory. There you will find several files demoing\nvarious parts of this library.\n\n## How it works?\n\nHere are all moving parts:\n\n1. **UserLike** object, aka "user model" - is an arbitrary class that implements `imia.UserLike` protocol.\n2. **a user provider** - an adapter that loads user model (UserLike object) from the storage (a database).\n3. **an authenticator** - a class that loads user using the user provider from the request (eg. session)\n4. **an authentication middleware** that accepts an HTTP request and calls authenticators for a user model. The\n   middleware always populates `request.auth` with `UserToken`.\n6. **user token** is a class that holds authentication state\n\nWhen a HTTP request reaches your application, an `imia.AuthenticationMiddleware` will start handling it. The middleware\niterates over configured authenticators and stops on the first one that returns non-None value. At this point the\nrequest is considered authenticated. If no authenticators return user model then the middleware will create  _anonymous\nuser token_. The user token available in `request.auth` property. Use `user_token.is_authenticated` token property to\nmake sure that user is authenticated.\n\n## User authentication quick start\n\n1. Create a user model and implement methods defined by `imia.UserLike` protocol.\n2. Create an instance of `imia.UserProvider` that corresponds to your user storage. Feel free to create your own.\n3. Setup one or more authenticators and pass them to the middleware\n4. Add `imia.AuthenticationMiddleware` to your Starlette application\n\nAt this point you are done.\n\nHere is a brief example that uses in-memory provider for demo purpose. For production environment you should use\ndatabase backed providers like `SQLAlchemyORMUserProvider` or  `SQLAlchemyCoreUserProvider`. Also, for simplicity reason\nwe will not implement [login/logout flow](docs/login_logout.md) and will authenticate requests using API keys.\n\n```python\nfrom dataclasses import dataclass, field\n\nfrom starlette.applications import Starlette\nfrom starlette.middleware import Middleware\nfrom starlette.requests import Request\nfrom starlette.responses import JSONResponse\nfrom starlette.routing import Route\n\nfrom imia import APIKeyAuthenticator, AuthenticationMiddleware, InMemoryProvider\n\n\n@dataclass\nclass User:\n    """This is our user model. It may be an ORM model, or any python class, the library does not care of it,\n    it only expects that the class has methods defined by the UserLike protocol."""\n\n    id: str\n    password: str = \'password\'\n    scopes: list[str] = field(default_factory=list)\n\n    def get_display_name(self) -> str:\n        return self.id.split(\'@\')[0].title()\n\n    def get_id(self) -> str:\n        return self.id\n\n    def get_hashed_password(self) -> str:\n        return self.password\n\n    def get_scopes(self) -> list:\n        return self.scopes\n\n\nasync def whoami_view(request: Request) -> JSONResponse:\n    return JSONResponse({\n        \'id\': request.auth.user_id,\n        \'name\': request.auth.display_name,\n    })\n\n\nuser_provider = InMemoryProvider({\n    \'user1@example.com\': User(id=\'user1@example.com\'),\n    \'user2@example.com\': User(id=\'user2@example.com\'),\n})\n\nauthenticators = [\n    APIKeyAuthenticator(user_provider=user_provider),\n]\n\nroutes = [\n    Route(\'/\', whoami_view),\n]\n\nmiddleware = [\n    Middleware(AuthenticationMiddleware, authenticators=authenticators)\n]\n\napp = Starlette(routes=routes, middleware=middleware)\n```\n\nNow save the file to `myapp.py` and run it with [uvicorn](https://uvicorn.org) application server:\n\n```bash\nuvicorn myapp:app\n```\n\nOpen `http://127.0.0.1:8000/` and see that your request is not authenticated and user is anonymous. Let\'s pass API key\nvia query parameters to make the configured APIKeyAuthenticator to load user. This time\nopen `http://127.0.0.1:8000/?apikey=user1@example.com` in your browser. Now the request is fully authenticated as User1\nuser.\n\nFor more details refer to the doc sections below.\n\n## Docs\n\n1. [UserLike protocol (a user model)](docs/userlike_protocol.md)\n2. [Load user from databases using User Providers](docs/user_providers.md)\n6. [Request authentication](docs/authentication.md)\n7. [Built-in authenticators](docs/authenticators.md)\n5. [User token](docs/user_token.md)\n5. [Passwords](docs/password_verification.md)\n4. [Login/Logout flow](docs/login_logout.md)\n8. [User impersontation](docs/impersonation.md)\n\n## Usage\n\nSee [examples/](examples) directory.\n',
    'author': 'Alex Oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/alex-oleshkevich/imia',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8.0,<4.0.0',
}


setup(**setup_kwargs)



from .authentication import *  # noqa
from .exceptions import *  # noqa
from .impersonation import *  # noqa
from .login import *  # noqa
from .protocols import *  # noqa
from .user_providers import *  # noqa
from .user_token import *  # noqa
